# Define high level functions for the project.
# Functions that are useful at high level are defined here.
# It include guessing content type, encoding and others.

from . import master
from . import document

__all__ = [
    "find_resource",
    "find_resources",
    "guess_content_type",
    "guess_encoding",
    "locally_hosted",
    "remotely_hosted",
    "is_supported",
    "is_resembled",
    "to_string",
    "guess_type",
    "guess_name",

    "is_path",
    "is_file_path",
    "is_dir_path",
    "is_path_like",
    "is_file_path_url",
    "is_url",
    "is_web_url",
    "is_file_like"
]

def find_resources(source, strict=False, **kwargs):
    # Find resources instances closely related to source.
    # master contains resource objects created from source.
    master_obj = master.Master(source, **kwargs)
    # Gets supported resources from master.
    # Supported resources clearly matches the source.
    # Supported resources take precidence over resembled ones.
    supported_resources = master_obj.supported_resources
    if supported_resources:
        return supported_resources
    elif not strict:
        # Gets resources resembling source.
        # Resemblimg resources are not guaranteed to be supported.
        # e.g url may still be invalid or file path not existing.
        return master_obj.resembles_resources
    else:
        return []
    
def find_resource(source, strict=False, **kwargs):
    # Find resource instance supporting or resembling source.
    # Returns None if source cannot satisfy any of resource object.
    # find_resources() but performance may be impacted.
    # Supported resource has priority over resmbling ones.
    master_obj = master.Master(source, **kwargs)
    supported_resource = master_obj.supported_resource
    if supported_resource:
        return supported_resource
    elif not strict:
        return master_obj.resembles_resource


def _find_resource_types(source, strict=False, **kwargs):
    # Returns resources classes supported or rembling source.
    # This is done by first finding resource objects and then getting
    # their classes.
    _resources = find_resources(source, strict, **kwargs)
    return [_resource.__class__ for _resource in _resources]

def _resource_access_attr(_resource, attribute, default=None):
    # Accesses attribute of resource object.
    # Takes into account resource being None(see find_resource())
    # 'default' argument will be returned in place of None.
    # If attribute does not exists exception will be raised.
    if _resource:
        value = getattr(_resource, attribute)
        if value != None:
            return value
    return default

def _find_resource_access_attr(source, attribute, strict=False, 
    default=None, **kwargs):
    # Finds resource object and access attribute from it.
    # None is returned when resource object cannot be found.
    # None also when the attribute has value of None(likely)
    resource_object = find_resource(source, strict, **kwargs)
    return _resource_access_attr(resource_object, attribute, default)

def guess_content_type(source):
    # Guesses content type of contents of source
    return _find_resource_access_attr(source, "content_type")

def guess_encoding(source):
    # Guesses encording of contents of source
    return _find_resource_access_attr(source, "encoding")

def locally_hosted(source):
    # Returns True if contents of source are available locally
    return _find_resource_access_attr(source, "locally")

def remotely_hosted(source):
    # Returns True if contents of source are available remotely.
    return _find_resource_access_attr(source, "remotely")

def is_supported(source):
    # Returns if source is supported(e.g is valid url)
    return _find_resource_access_attr(source, "supported", default=False)

def is_resembled(source):
    # Returns True if source resembles classes for supported sources.
    # Source may be valid file path but not existing.
     return _find_resource_access_attr(source, "resembles", default=False)

def to_string(source):
    # returns string version of source(None if not supported)
    return _find_resource_access_attr(source, "uri")

def guess_type(source):
    # Returns string about type of source provided source.
    # It may be files path, file like object, etc.
    return _find_resource_access_attr(source, "uri_type")

def guess_name(source):
    # Guesses name of source(string version of source)
    return to_string(source)



def _source_matches_resource_type(source, resource_type, strict=False):
    # Checks if ssource upports or resembles resource type(class)
    # resource_types = _find_resource_types(source, strict)
    # return resource_type in resource_types
    _resource = resource_type(source)
    supported = _resource.supported
    if supported:
        return supported
    elif not strict:
        return _resource.resembles
    else:
        return False

def is_path(source, strict=False):
    # Checks if source is dir or file path
    return _source_matches_resource_type(source, document.Path, strict)

def is_file_path(source, strict=False):
    # Checks if source is file path
    return _source_matches_resource_type(source, document.FilePath, strict)

def is_dir_path(source, strict=False):
    # Checks if source is file path
    return _source_matches_resource_type(source, document.DirPath, strict)

def is_path_like(source, strict=False):
    # Checks if source is a path like object
    return _source_matches_resource_type(source, document.PathLike, strict)

def is_file_path_url(source, strict=False):
    # Checks if source is url for local file path(file://)
    return _source_matches_resource_type(source, document.FilePathURL, strict)

def is_url(source, strict=False):
    # Checks if source is url(file path not included)
    return _source_matches_resource_type(source, document.URL, strict)

def is_web_url(source, strict=False):
    # Checks if source is url for web documents
    return _source_matches_resource_type(source, document.WebURL, strict)

def is_file_like(source, strict=False):
    # Checks if source is file like object
    return _source_matches_resource_type(source, document.FileMemory, strict)

def is_memory_file(source, strict=False):
    # Checks if source is file like object
    return is_file_like(source, strict)



if __name__ == "__main__":
    resource = find_resource("file.py")
    print(guess_type("git://www.example.com/search?q=mimetypes+"))
    


from . import pathmod
from . import filepath
from . import file_memory

from . import urlmod
from . import weburl

from . import resource
from . import exceptions

import os


# __all__ = [
#     "Document",
#     "WebURL",
#     "FilePath",
#     "FilePathLike",
#     "FilePathURL",
#     "FileMemory"
# ]

class Document(resource.Resource): 
    _path_types = pathmod.PATH_TYPES

    def __init__(self, source, content_type=None, encoding=None):
        super().__init__(source, content_type, encoding)

    def source_supported(self, source):
        # Checks if source is supported
        # Any object is supported and can be used as source
        return isinstance(source, object)

    def to_string(self):
        # Returns string version of source.
        # source or path will be used if they are strings.
        if self._source and isinstance(self._source, str):
            return self._source
        elif self.path and isinstance(self.path, str):
            return self.path
        else:
            return super().to_string()

    def extract_path(self, source):
        # Extracts path from source
        return ""

    @property
    def path(self):
        return self.extract_path(self._source)

    @property
    def content_type(self):
        if self.path:
            return filepath.guess_content_type(self.path)

    @property
    def encoding(self):
        if super().encoding:
            return super().encoding
        elif self.content_type:
            # Guessses encoding from existing content type
            extension = filepath.guess_extension(self.content_type)
            if extension:
                return filepath.guess_encoding(" " + extension)


class URL(Document):
    _uri_type = "url"

    def extract_path(self, url):
        # Extracts path from source
        return urlmod.extract_path(self._source)

    def source_supported(self, source):
        # Checks if source is valid url
        return urlmod.is_url(source)

    def source_resembles(self, source):
        # Checks if source resembles url
        return urlmod.resembles_url(source)

    def available_locally(self, url):
        return urlmod.is_locally_hosted(url)

    @property
    def hostname(self):
        return urlmod.extract_hostname(self._source)

    @property
    def netloc(self):
        return urlmod.extract_netloc(self._source)

    @property
    def scheme(self):
        return urlmod.extract_scheme(self._source)

    @property
    def query(self):
        return urlmod.extract_query(self._source)

    @property
    def params(self):
        return urlmod.extract_params(self._source)

    @property
    def fragment(self):
        return urlmod.extract_fragment(self._source)

    @property
    def port(self):
        return urlmod.extract_port(self._source)


class WebURL(URL):
    _uri_type = "web-url"

    @property
    def content_type(self):
        # Retrieves content type for contents of url
        if super().content_type:
            return super().content_type
        else:
            # html file is basic structure for webpage.
            # content type for webpage is same as that of html
            return filepath.guess_content_type(" .html")

    def source_supported(self, source):
        # Checks if source is valid url for web
        return weburl.is_web_url(source)

    def source_resembles(self, source):
        # Checks if source resembles url
        return weburl.resembles_web_url(source)

    @property
    def is_webpage(self):
        # Checks if source(url) points to webpage
        html_content_type = filepath.guess_content_type(" .html")
        return html_content_type == self.content_type
 


class Path(Document):
    _uri_type = "file-system-path"

    def __init__(self, source, content_type=None, encoding=None):
        super().__init__(source, content_type, encoding)
    
    def extract_path(self, path):
        if isinstance(path, os.PathLike):
            return path.name
        else:
            return path

    def source_supported(self, source):
        return pathmod.is_path(source, strict=True)

    def available_locally(self, path):
        return True

    def source_resembles(self, source):
        return pathmod.resembles_path(source)

    @property
    def size(self):
        return os.stat(self.path).st_size

    @property
    def access_time(self):
        return os.stat(self.path).st_atime

    @property
    def mod_time(self):
        return os.stat(self.path).st_mtime

    @property
    def create_time(self):
        return os.stat(self.path).st_ctime


class DirPath(Path):
    _uri_type = "dir-path"

    def source_supported(self, source):
        return pathmod.is_dir_path(source, strict=True)

    def source_resembles(self, source):
        return pathmod.resembles_dir(source)

    def get_files(self, recursive=False):
        return pathmod.get_folder_files(self.path, recursive)

    def get_dirs(self, recursive=False):
        return pathmod.get_folder_dirs(self.path, recursive)

    @property
    def files(self):
        return self.get_files(False)

    @property
    def dirs(self):
        return self.get_dirs(False)

    @property
    def dirs_recursive(self):
        return self.get_dirs(True)
    
    @property
    def files_recursive(self):
        return self.get_files(True)

class FilePath(Path):
    _uri_type = "file-path"
    def source_supported(self, source):
        # File path is expected to be string or bytes.
        # Pathlike, integers cant be used as path.
        return pathmod.is_file_path(source, strict=True)

    def source_resembles(self, source):
        # Checks if source resembles path
        return pathmod.resembles_file_path(source)

class PathLike(Path):
    _uri_type = "path-like-object"
    def extract_path(self, file_like):
        return file_like.name

    def source_supported(self, file_like):
        if isinstance(file_like, os.PathLike):
            file_path = self.extract_path(file_like)
            return super().source_supported(file_path)
        else:
            return False

    def source_resembles(self, source):
        return isinstance(source, os.PathLike)

class DirPathLike(PathLike):
    _uri_type = "dir-path-like-object"
    pass

class FilePathLike(PathLike):
    _uri_type = "file-path-like-object"
    pass

    

class FilePathURL(URL, FilePath):
    _uri_type = "local-file-url"
    def extract_path(self, file_path):
        path_part = urlmod.extract_path(file_path)
        # Decodes extracted path from url
        path_decoded = urlmod.unquote(path_part)
        # Removes leading(/) if path has drive part
        if pathmod.extract_drive(path_decoded[1:]):
            return path_decoded[1:]
        return path_decoded


    def source_supported(self, source):
        return urlmod.is_url(source, {"file"})

    def source_resembles(self, source):
        return urlmod.resembles_url(source, {"file"})

    def available_locally(self, file):
        return True
 
class FileMemory(Document):
    _uri_type = "file-like-object"
    def set_path(self, path):
        # Sets associated with source
        if isinstance(path, pathmod.PATH_TYPES):
            self.path = path
        else:
            err_msg = "Path should be str, bytes or os.PathLike not {}"
            raise TypeError(err_msg.format(type(path)))


    def extract_path(self, file):
        # Extracts path from 'name' attribute of file object.
        # Empty string will be returned if 'name' attributes not exists.
        try:
            name = file.name
        except AttributeError:
            pass
        else:
            if isinstance(name, pathmod.PATH_TYPES):
                return name
    
    def source_supported(self, source):
        # Checks source satisfies requirements for file object.
        # 'name' attribute is expected to be valid path.
        return file_memory.is_memory_file(source)

    def available_locally(self, file):
        return True


if __name__ == "__main__":
    import os
    import tempfile
    import pathlib

    _file_path = "https://www.example.com/search.filenames"
    with tempfile.TemporaryFile() as f:
        file = Path("setup\\")
        print(file.path)
        #file.set_path("filename.pdf")
        print(file.content_type, file.status)
            


import os
import glob
import typing
from urllib import parse

import tempfile

# Defines allowed types for paths
PATH_TYPES = (str, bytes, os.PathLike)
_PATH = typing.Union[str, bytes, os.PathLike]

# Max size of characters pathname(e.g filename) can have.
# This does not mean the whole path.
MAX_PATH_NAME_LENGTH = 100


def walk(path: _PATH, recursive=False):
    # Returns paths(dirs, files) in provided path.
    # recursive if True match paths in sub-folder recursively.
    for root, directories, filenames in os.walk(path):
        for pathname in directories + filenames:
            yield os.path.join(root, pathname)
        if not recursive:
            break

def filter_files(paths: typing.Iterable[_PATH]):
    return filter(os.path.isfile, paths)

def filter_dirs(paths: typing.Iterable[_PATH]):
    return filter(os.path.isdir, paths)


def get_folder_files(path: _PATH, recursive=False):
    # Returns file paths in folder
    return filter_files(walk(path, recursive))

def get_folder_dirs(path: _PATH, recursive=False):
    # Returns file paths in folder
    return filter_dirs(walk(path, recursive)) 

def glob_pattern_paths(glob_path, recursive=False):
    return glob.glob(glob_path, recursive=recursive)


def extract_drive(path):
    return os.path.splitdrive(path)[0]

def remove_drive(path):
    return os.path.splitdrive(path)[1]


def is_path(_source, exists_callback=None, strict=True):
    if exists_callback == None:
        exists_callback = os.path.exists
    if isinstance(_source, PATH_TYPES):
        if strict:
            return exists_callback(_source)
        else:
            if exists_callback(_source):
                return True
            else:
                try:
                    with tempfile.NamedTemporaryFile('w', prefix=_source):
                        pass
                except OSError:
                    return False
                else:
                    return True
    else:
        return False


def is_file_path(_source, strict=True):
    return is_path(_source, os.path.isfile, strict)

def is_dir_path(_source, strict=True):
    return is_path(_source, os.path.isdir, strict)

def resembles_path(_source):
    # Guesses if object resembles path
    if _source and isinstance(_source, PATH_TYPES):
        if isinstance(_source, os.PathLike):
            return True
        _source = str(_source)
        drive, path = os.path.splitdrive(_source)
        parsed = parse.urlparse(_source)

        if parsed.netloc or parsed.params:
            # Path cant have hostname(likely other urls)
            return False
        elif parsed.scheme and not drive:
            # Path cant have scheme but miss drive part.
            return False
        else:
            path = os.path.normpath(path)
            path_names = filter(None, path.split(os.sep))
            for path_name in path_names:
                if not is_path(path_name, os.path.exists, False):
                    return False
            return True


def resembles_dir(_source):
    # Guesses if object resembles directory path
    if resembles_path(_source):
        # dir should not have to end with path sep or extension.
        path = os.path.normcase(_source)
        extension = os.path.splitext(path)[1]
        return path.endswith(os.sep) or not extension
    else:
        return False

def resembles_file_path(_source):
    # Guesses if object resembles file path
    if resembles_path(_source):
        #extension = os.path.splitext(_source)[1]
        path = os.path.normcase(_source)
        return not path.endswith(os.sep)
    else:
        return False


if __name__ == "__main__":
    import pathlib
    path = "/stackoverflow.com/\questions/3207219\\"
    print(resembles_file_path(path))


from . import filepath
from . import weburl
from . import urlmod

def is_webpage(_source):
    # Checks if source is points to webpage
    if weburl.is_web_url(_source):
        path = urlmod.extract_path(_source)
        if path:
            if filepath.extract_extension(path):
                # url resembling html is likely webpage
                # e.g https://www.example.com/sample.html
                return filepath.is_html_file(path)
            else:
                # url has path part but less likely to be a file.
                # e.g https://www.example.com/page
                return True
        else:
            # No path part means its likely webpage.
            # e.g https://www.example.com/
            return True
    else:
        return False
        


import io


def is_memory_file(_source):
    # Checks if source is a file like object
    # created from: navaly source/naval/utility/file.py
    if isinstance(_source, io.IOBase):
        return True
    elif hasattr(_source, "write"):
        file_objs_attr = ["read", "truncate", "seek", "closed"]
        # this will work for _TemporaryFileWrapper class
        # its returned by tempfile.TemporaryFile() on windows
        return all([hasattr(_source, attr) for attr in file_objs_attr])
    else:
        return False



from . import urlmod
from . import weburl
from . import filepath
from . import file_memory


def is_memory_file(_source):
    # Checks if source is a file like object
    # created from: navaly source/naval/utility/file.py
    return file_memory.is_memory_file(_source)

def is_remote_file(_source):
    # Checks if source points to remote file(e.g webpage)
    if weburl.is_web_url(_source):
        # extract the path part of url
        path_part = urlmod.extract_path(_source)
        content_type = filepath.guess_content_type(path_part)
        # url pointing to file should have file extension
        return bool(content_type)
    return False


def is_local_file(_source):
    # Checks if source points to file on local machine
    if not is_remote_file(_source):
        return filepath.is_file_path(_source, False)
    else:
        return False


     



# Manages resource classes and objects.
# It takes care of creating resource object based on source.
# This is inspired from 'navaly project.

from . import resource
from . import document

from typing import List
from typing import Type
from typing import Iterable


# Default resources classes(order matters)
# The more specific class is the more top it is.
# FilePath is more specific than Path.
_DEFAULT_RESOURCE_TYPES = (
    # file-system like path
    document.FilePathURL,
    document.FilePathLike,
    document.FilePath,
    document.DirPathLike,
    document.DirPath,
    document.Path,

    # urls
    document.WebURL,
    document.URL,

    # file like object
    document.FileMemory,
)
resources_types = [*_DEFAULT_RESOURCE_TYPES]



class ResourceTypes():
    # Manages Resource types(classes) to be used by Master class.
    # Class is not used in favour of list of Resource objects.
    def __init__(self, resources_types) -> None:
        self._resource_types = list(resources_types)

    def add_resource_type(self, resource_type):
        self._resource_types.add(resource_type)

    def remove_resource_type_(self, resource_type):
        self._resource_types.discard(resource_type)

    def get_resource_types(self):
        return self._resource_types


class Resources():
    # Manages Resource instances to be used by Master class.
    def __init__(self, resources: Iterable[resource.Resource]) -> None:
        self._resources = resources

    def set_resources(self, resources):
        self._resources = resources

    def get_resources(self):
        return self._resources

    def _filter(self, resources, callback) -> List[resource.Resource]:
        # returns list with filtered resources by provided callback
        return list(filter(callback, resources))

    def _single_filter(self, resources, callback) -> resource.Resource:
        # Returns first resource satisfyting callback
        for resource in resources:
            if callback(resource):
                return resource
        return None

    @property
    def supported_resources(self):
        # Returns list of supported resources
        return self._filter(self._resources, lambda res: res.supported)

    @property
    def resembles_resources(self):
        # Returns list of resources resembling source
        return self._filter(self._resources, lambda res: res.resembles)

    @property
    def issues_resources(self):
        # Returns list of resources with issues
        return self._filter(self._resources, lambda res: res.issues)

    @property
    def success_resources(self):
        # Returns list of success resources.
        # This is opposite of issues_resources.
        return self._filter(self._resources, lambda res: res.success)


    @property
    def supported_resource(self):
        # Returns first resource supported.
        callback = lambda res: res.supported
        return self._single_filter(self._resources, callback)

    @property
    def resembles_resource(self):
        # Returns first resource resembling its source
        callback = lambda res: res.resembles
        return self._single_filter(self._resources, callback)

    @property
    def issues_resource(self):
        # Returns first resource with issues
        callback = lambda res: res.issues
        return self._single_filter(self._resources, callback)

    @property
    def success_resource(self):
        # Returns first success resource.
        callback = lambda res: res.success
        return self._single_filter(self._resources, callback)


class Master(Resources):
    # Master class for Resource classes.
    # Creates Resource instances on request and help in accessing them.
    # Internally creates instances as compared to parent class Resources.
    def __init__(self, source, *args, **kwargs):
        super().__init__(list())
        self._source = source
        self._args = args
        self._kwargs = kwargs
        # Calling object method within initialiser can cause problems.
        # Object not yet completely created.
        self._resource_types = resources_types.copy()
        # Updates resources instances as resources types(classes) changed.
        # set_resource_types() automatically does update resource objects.
        self.__update_resources()


    def set_resource_types(self, resource_types):
        # Sets resources types(classes) and updates instances.
        self._resource_types = resource_types
        self.__update_resources()

    def get_resources_types(self):
        return self._resource_types


    def __create_resources(self):
        # Creates resources objects for source using resources classes.
        def func(resource_type):
            return resource_type(self._source, *self._args, **self._kwargs)
        # Convert to list to avoid iterator from getting exausted.
        # That is common with python iterator
        return list(map(func, self._resource_types))

    def __update_resources(self):
        # Updates resource resources by creating new instances.
        self._resources = self.__create_resources()


if __name__ == "__main__":
    source = "file"
    master = Master(source)
    print(master.resembles_resource.uri_type)
    


from . import exceptions


class Resource(): 
    # Base class for analysing source for information like content type.
    # Most of attributes can be accessed as properties.
    # It contains properties for determining if source is supported,
    # located locally and others.
    # This is the most important class in this library.
    _uri_type = None

    def __init__(self, source, content_type:str=None, encoding:str=None):
        self._source = source
        self._content_type = content_type
        self._encoding = encoding


    def _validate_source(self):
        # Raises exception when source is not supported
        if not self.source_supported(self._source):
            err_msg = "Source '{}' is not supported for {}"
            err_msg = err_msg.format(
                self.uri[:20], 
                type(self)
            )
            raise exceptions.UnsupportedSourceError(err_msg)

    def source_supported(self, source):
        # Checks if source is supported
        # Any object is supported and can be used as source.
        # No source is supported for now.
        return False

    def to_string(self):
        # Returns string version of source
        return ""

    def to_dict(self):
        return {
            "uri": self.uri,
            "uri_type": self.uri_type,
            "content_type": self.content_type,
            "encoding": self.encoding,
            "supported": self.supported,
            "resembles": self.resembles,
            "available_localy": self.locally,
            "issues": self.issues,
            "status": self.status
        }

    def available_locally(self, source):
        # Checks if source data is available on this local machine
        return False

    def source_resembles(self, source):
        # Returns True if source is likely to be supported.
        # That does not mean source is supported.
        return self.source_supported(source)

    @property
    def supported(self):
        return self.source_supported(self._source)

    @property
    def resembles(self):
        return self.source_resembles(self._source)

    @property
    def issues(self):
        return not (self.supported or self.resembles)

    @property
    def success(self):
        return not self.issues

    @property
    def status(self):
        if self.supported:
            return "OK"
        elif self.source_resembles(self._source):
            return "Not Supported(resembles)"
        else:
            return "Not Supported(critically!!)"

    @property
    def locally(self):
        return self.available_locally(self._source)

    @property
    def remotely(self):
        return not self.available_locally(self._source)

    @property
    def uri(self):
        return self.to_string()

    @property
    def uri_type(self):
        return self._uri_type

    @property
    def summary(self):
        return self.to_dict()
    
    @property
    def source(self):
        return self._source

    @property
    def content_type(self):
        return self._content_type

    @property
    def encoding(self):
        return self._encoding


class Generic(Resource):
    # This sources that are generic but useful.
    # This can include path which can be dir path or url path.
    # Or url which can be web url, local file url, etc.
    # Sources for in this class do
    def source_supported(self, source):
        # Does not need to be supported as its too generic
        return False

    @property
    def supported(self):
        # Does not need to be supported as its too generic
        return False



from . import exceptions

import socket
from urllib import parse


LOCAL_HOST_NAMES = {"localhost", "127.0.0.1"}

# url parsing functions
def urlparse(url):
    return parse.urlparse(url)

def urlparse_dict(url):
    return urlparse(url)._asdict()

def url_unparse(url):
    return parse.urlparse(url)._asdict()

def urljoin(base_url, url):
    return parse.urljoin(base_url, url)

def quote(url, *args, **kwargs):
    return parse.quote(url, *args, **kwargs)

def unquote(url, *args, **kwargs):
    return parse.unquote(url, *args, **kwargs)


# Functions for extracting parts of url
def extract_path(url: str):
    # Extracts path part of url
    return parse.urlparse(url).path

def extract_hostname(url: str):
    # Extracts hostname part of url
    return parse.urlparse(url).hostname

def extract_netloc(url: str):
    # Extracts netloc part of url
    return parse.urlparse(url).netloc

def extract_scheme(url: str):
    # Extracts scheme of url
    return parse.urlparse(url).scheme

def extract_params(url: str):
    # Extracts parameter passed to url
    return parse.urlparse(url).params

def extract_query(url: str):
    return parse.urlparse(url).query

def extract_port(url: str):
    port = parse.urlparse(url).port
    if port == None:
        return ""
    else:
        return str(port)

def extract_fragment(url: str):
    return parse.urlparse(url).fragment


# # Functions for validationg url
# def is_file_path_scheme(scheme):
#     # Checks if chememe is that of file path.
#     # credit: https://stackoverflow.com/questions/29026051
#     return os.path.exists(scheme + ":\\")


def is_url(_source, schemes=None):
    # Checks if source is url for web resource.
    if isinstance(_source, (str, bytes)):
        scheme = extract_scheme(_source)
        netloc = extract_netloc(_source)
        path = extract_path(_source)
        any_items = (netloc, path)
        if schemes != None:
            return scheme in schemes and any(any_items)
        else:
            return bool((scheme)) and any(any_items)
    return False

def resembles_scheme(_url, scheme):
    # Checks if url resembles containing scheme
    # It does not mean that scheme is real scheme of url.
    url_scheme = extract_scheme(_url)
    if scheme != None:
        return scheme == url_scheme
    else:
        return bool(url_scheme)
        


def resembles_hostname(_url, hostname):
    # Checks if url resembles containing hostname
    url_hostname = extract_hostname(_url)
    if hostname != None:
        return url_hostname == hostname
    else:
        return bool(url_hostname)

def resembles_schemes(_url, schemes):
    # Checks if url resembles containing schemes
    for scheme in schemes:
        if resembles_scheme(_url, scheme):
            return True
    return False

def resembles_url(_source, schemes=None):
    # Checks if source resembles url.
    # 1. source should 2 of any of url parts excluding path and hostname.
    # 2. Or source should have scheme(or resemble) and path.
    # 2. Or source should have atleast hostname.
    if isinstance(_source, (str, bytes)):
        if schemes != None:
            # URL needs to have atleast one of provided schemes
            scheme_satisfied = resembles_schemes(_source, schemes)
        else:
            # Dont care if url has scheme if schemes not provided
            scheme_satisfied = True

        if scheme_satisfied and resembles_hostname(_source, None):
            # Hostname is enough to satify url
            return True
        elif scheme_satisfied and extract_path(_source):
            # scheme and path satisfies url
            return True
        else:
            # source satifying scheme and atleast 2 other parts of url
            # can be considered resembling url.
            any_items = [
                extract_query(_source),
                extract_params(_source),
                extract_port(_source),
                extract_fragment(_source)
            ]
            any_items_filtered = list(filter(None, any_items))
            return scheme_satisfied and len(any_items_filtered) >= 2
    return False


# Functions for transforming url
def make_url_absolute(url:str, base_url:str):
    # Makes url absolute by adding missing parts from base_url
    # inspired by: requests-html requests_html.BaseParse._make_absolute()

    # Parse url componets into dictionary
    parsed = urlparse(url)._asdict()

    # url almost complete but missing scheme
    if url.startswith("//"):
        parsed['scheme'] = urlparse(base_url).scheme

        # Recreates url with new scheme
        parsed = [value for value in parsed.values()]
        return url_unparse(parsed)

    # Link is absolute; its just missing scheme and netloc
    elif url.startswith("/"):
        parsed['scheme'] = urlparse(base_url).scheme
        parsed['netloc'] = urlparse(base_url).netloc

        # Recreates url with new scheme and netloc
        parsed = [value for value in parsed.values()]
        return url_unparse(parsed)

    # The link is relative to current url
    elif not parsed['netloc']:
        # parse.urljoin() is intelligent than os.path.join()
        return urljoin(base_url, url)
    
    # URL is complete and should have everything
    elif is_url(url):
        return url
    else:
        raise exceptions.InvalidURLError(url + " is not a valid url")



# Functions relating to IP addresses
def _get_local_adress():
    # Gets ip adress of local machine
    hostname = socket.gethostname()
    return socket.gethostbyname(hostname)

def _get_local_ip_adresses():
    # Gets ip adresses of local machine.
    # Functions is not reliable and shouldnt be depended.
    # source: https://stackoverflow.com/questions/270745/
    addreses = set()
    for info in socket.getaddrinfo(socket.gethostname(), None):
        addreses.add(info[4][0])
    return addreses

def is_local_ip_address(ip_address):
    # Checks if url is built on local ip address.
    # Not reliable function.
    return ip_address in _get_local_ip_adresses()


# Functions for checking if url is hosted locally or remotely
def is_local_host(hostname):
    # Checks hostname identifies this machine e.g localhost
    # Not reliable function.
    return hostname in LOCAL_HOST_NAMES

def is_remote_host(hostname):
    # Checks if host is remote host e.g example.com
    return not is_local_host(hostname)

def is_locally_hosted(_url):
    # Checks if url is served by local machine.
    # Functions is not reliable and shouldnt be depended.
    hostname = extract_hostname(_url)
    if hostname:
        return is_local_host(hostname) or is_local_ip_address(hostname)
    else:
        return True

def is_remotely_hosted(_url):
    # Checks if url is served by remote machine.
    # Functions is not reliable and shouldnt be depended.
    return not is_locally_hosted(_url)


if __name__ == "__main__":
    base_url = "parsy.readthedocs.io/en/latest?name#10/name"
    relative_url = "tutorial.html"
    print(resembles_url(base_url, None))
    print(make_url_absolute(base_url, base_url))


from . import urlmod


WEB_URL_SCHEMES = {"http", "https", "ftp"}

def is_web_url(_source):
    # Checks if source is url for web resource
    return urlmod.is_url(_source, WEB_URL_SCHEMES)

def resembles_web_url(_source):
    return urlmod.resembles_url(_source, WEB_URL_SCHEMES)


if __name__ == "__main__":
    base_url = "https://parsy.readthedocs.io/en/latest/"
    print(is_web_url(base_url))
    print(resembles_web_url("http://172.18.128.1/"))





class ResidError(Exception):
    '''Base exception class for all resid errors'''
    pass

class SourceError(ResidError):
    pass

class InvalidURLError(SourceError):
    pass

class FIlePathError(SourceError):
    pass

class InvalidFilePathError(FIlePathError):
    pass

class UnsupportedSourceError(SourceError):
    pass



import os
import mimetypes
import typing

from . import pathmod


if not mimetypes.inited:
    mimetypes.init()

def _bytes_to_string(__object, *args, **kwargs):
    # Convert bytes object to string.
    # Other object types are returned unchanged.
    if isinstance(__object, bytes):
        return __object.decode(*args, **kwargs)
    else:
        return __object

def guess_content_type(file_path):
    file_path = _bytes_to_string(file_path)
    return mimetypes.guess_type(file_path, False)[0]

def guess_encoding(file_path):
    file_path = _bytes_to_string(file_path)
    return mimetypes.guess_type(file_path, False)[1]

def guess_extension(content_type):
    return mimetypes.guess_extension(content_type, False)

def guess_extensions(content_type):
    file_path = _bytes_to_string(file_path)
    return mimetypes.guess_all_extensions(content_type, False)

def equal_content_type(file_path, other_file_path):
    # Returns True if files paths have same content type
    file_path_type = guess_content_type(file_path)
    other_file_path_type = guess_content_type(other_file_path)
    if file_path_type and other_file_path_type:
        return file_path_type == other_file_path_type
    else:
        return False


def extract_extension(file_path):
    return os.path.splitext(file_path)[1]

def extract_filename(file_path):
    return os.path.split(file_path)[1]

def remove_extension(filename: str):
    return os.path.splitext(filename)[0]


    
def is_pdf_file(file_path):
    return equal_content_type(file_path, " .pdf")

def is_docx_file(file_path):
    return equal_content_type(file_path, " .docx")

def is_doc_file(file_path):
    return equal_content_type(file_path, " .doc")

def is_word_file(file_path):
    return is_docx_file(file_path) or is_doc_file(file_path)

def is_pptx_file(file_path):
    return equal_content_type(file_path, " .pptx")

def is_ppt_file(file_path):
    return equal_content_type(file_path, " .ppt")

def is_html_file(file_path):
    if equal_content_type(file_path, " .html"):
        return True
    elif equal_content_type(file_path, " .htm"):
        return True
    else:
        return False

def is_json_file(file_path):
    return equal_content_type(file_path, " .json")

def is_csv_file(file_path):
    return equal_content_type(file_path, " .csv")

def is_plain_text_file(file_path):
    return equal_content_type(file_path, " .txt")

def is_text_file(file_path):
    content_type = guess_content_type(file_path)
    if content_type:
        return "text/" in content_type
    else:
        return False


if __name__ == "__main__":
    path = "/projects/resid.com"
